import os
import re
import json
from dateutil import parser


def convert_benchmark(output_dir, monitoring_log):
    """
    converts and Saves a downloaded monitoring log to a formated JSON
    The monitoring script can be found here gs://broad-gdr-encode-caas-execution/scripts/monitoring.sh
    :param output_dir: the desired output directory to store the combined monitoring log json file
    :param monitoring_log: monitoring log downloaded from the googgle bucket (JSON)
    """

    with open(monitoring_log, "r") as bm:
        # read json in a dict and
        call_logs = json.load(bm)

    # for each task call in dict
    converted_call_logs = {}

    for name, info in call_logs.items():
        # summary
        num_of_cores = None
        total_mem = {"size": None, "unit": None}
        total_disk_sapce = {"size": None, "unit": None}

        # log info
        logs = []
        time_stamp = None
        cpu_usage = None
        mem_usage = None
        disk_usage = None

        # for each line in the task call info
        lines = info.split("\n")
        for line in lines:
            # if a line contains the desired info
            if '#CPU:' in line:
                # get that info
                info = get_info("\d+", line)

                # if the info is of correct format
                if info is not None:
                    # save that info
                    num_of_cores = int(info[0])

            elif 'Total Memory:' in line:
                # $ for matching at the end of the string
                info = get_info("(\d*\.\d+|\d+)(M|G|T)$", line)

                if info is not None:
                    total_mem = {"size": float(info[0]), "unit": str(info[1]) + "B"}

            elif 'Total Disk space: ' in line:
                info = get_info("(\d*\.\d+|\d+)(M|G|T)$", line)

                if info is not None:
                    total_disk_sapce = {
                        "size": float(info[0]),
                        "unit": str(info[1]) + "B",
                    }

            elif '[' in line:
                time_stamp = str(
                    parser.parse(line.replace("[", '').replace("]", '').strip())
                )

            elif "CPU usage: " in line:
                info = get_info("(\d*\.\d+|\d+)(%)$", line)

                if info is not None:
                    cpu_usage = float(info[0])

            elif '* Memory usage: ' in line:
                info = get_info("(\d*\.\d+|\d+)(%)$", line)

                if info is not None:
                    mem_usage = float(info[0])

            # if you have read disk usage, one time stamp has been read so save that information in the log
            elif '* Disk usage: ' in line:
                info = get_info("(\d*\.\d+|\d+)(%)$", line)

                if info is not None:
                    disk_usage = float(info[0])

                log = {
                    "time": time_stamp,
                    "cpu_usage": cpu_usage,
                    "memory_usage": mem_usage,
                    "disk_usage": disk_usage,
                }
                logs.append(log)

                disk_usage = None
                time_stamp = None
                cpu_usage = None
                mem_usage = None

        # add the task call name, summary and logs to a dict
        call_log = {"id": name}
        summary = {
            "cores": num_of_cores,
            "memory": total_mem,
            "disk_space": total_disk_sapce,
        }

        call_log["summary"] = summary
        call_log["logs"] = logs
        converted_call_logs.setdefault("calls", []).append(call_log)

    name = os.path.splitext(os.path.basename(monitoring_log))[0]
    file_name = get_path(name + "_formatted.json", output_dir)
    store_benchmark(converted_call_logs, file_name)


def get_info(prefix, string):
    """
    :param prefix: the regex to match the info you are trying to obtain
    :param string: the string where the info is contained (can have new line character)
    :return: the matches within the line
    """

    info = None

    # find and return the matches based on the prefix and if there is a match (not empty)
    matches = re.findall(prefix, string)
    if len(matches) > 0:

        info = matches[0]

    return info


def store_benchmark(call_logs, file_name):
    """
    dumps the information of each call task in dictionary format into one JSON file
    :param call_logs: the information of each call task in dictionary format
    :param file_name: name of the file with absolute path to dump the dictionay into
    """

    with open(file_name, "w") as logs:

        json.dump(call_logs, logs, sort_keys=False, indent=4)


def get_path(filename, output_dir):
    """
    :param filename: the filename to be saved in the output directory
    :param output_dir: the desired output directory to store the monitoring log, plots and json file
    :return: returns the combined path of file name and  output directory
    """

    # get the combined path of output dir path, filename and sub dir path
    path = os.path.join(output_dir, filename)

    # if the combined path exist
    if os.path.isfile(path):

        # delete that current path
        os.remove(path)

    # return combined path
    return path



"""
TODO: add some module docs
TODO: once switched to support only Py3.7+, replace all 'cls'
type annotations with the actual Types, rather than using the strings.
This in Py3.6(-) is limited by the lack of Postponed Evaluation of Annotations, see:
https://www.python.org/dev/peps/pep-0563/
"""

import time

import io
import json
import logging
import requests
from datetime import datetime, timedelta

from cromwell_tools.cromwell_auth import CromwellAuth
from cromwell_tools import utilities
from cromwell_tools.utilities import validate_cromwell_label
from cromwell_tools import exceptions
from typing import List, Union, Dict


logger = logging.getLogger(__name__)

_failed_statuses = ('Failed', 'Aborted', 'Aborting')

_cromwell_exclusive_query_keys = {
    'end',
    'includeSubworkflows',
    'start',
    'submission',
    'page',
    'pageSize',
}

_cromwell_inclusive_query_keys = {
    'additionalQueryResultFields',
    'excludeLabelAnd',
    'excludeLabelOr',
    'id',
    'includeSubworkflows',
    'label',
    'labelor',
    'name',
    'status',
}

_cromwell_query_keys = _cromwell_exclusive_query_keys.union(
    _cromwell_inclusive_query_keys
)


# TODO: use functools partial for get, post (set the authenticate commands)
class CromwellAPI(object):
    """Contains a set of classmethods that implement interfaces to cromwell REST API endpoints."""

    # TODO: move the endpoints definitions to the corresponding functions after refactoring the unit tests and mocks
    _abort_endpoint = '/api/workflows/v1/{uuid}/abort'
    _status_endpoint = '/api/workflows/v1/{uuid}/status'
    _submit_endpoint = '/api/workflows/v1'
    _metadata_endpoint = '/api/workflows/v1/{uuid}/metadata'
    _health_endpoint = '/engine/v1/status'
    _release_hold_endpoint = '/api/workflows/v1/{uuid}/releaseHold'
    _query_endpoint = '/api/workflows/v1/query'

    @classmethod
    def abort(
        cls: 'CromwellAPI',
        uuid: str,
        auth: CromwellAuth,
        raise_for_status: bool = False,
    ) -> requests.Response:
        """Request Cromwell to abort a running workflow by UUID.

        Args:
            uuid: A Cromwell workflow UUID, which is the workflow identifier.
            auth: The authentication class holding headers or auth
                information to a Cromwell server.
            raise_for_status: Whether to check and raise for status based on the response.

        Raises:
            requests.exceptions.HTTPError: This will be raised when raise_for_status is True and Cromwell returns
                a response that satisfies 400 <= response.status_code < 600.

        Returns:
            HTTP response from Cromwell.
        """
        response = requests.post(
            url=auth.url + cls._abort_endpoint.format(uuid=uuid),
            auth=auth.auth,
            headers=auth.header,
        )

        if raise_for_status:
            cls._check_and_raise_status(response)
        return response

    @classmethod
    def metadata(
        cls: 'CromwellAPI',
        uuid: str,
        auth: CromwellAuth,
        includeKey: Union[List[str], str] = None,
        excludeKey: Union[List[str], str] = None,
        expandSubWorkflows: bool = False,
        raise_for_status: bool = False,
    ) -> requests.Response:
        """Retrieve the workflow and call-level metadata for a specified workflow by UUID.

        Args:
            uuid: A Cromwell workflow UUID, which is the workflow identifier.
            auth: The authentication class holding headers or auth
                information to a Cromwell server.
            includeKey: When specified key(s) to include from the metadata. Matches any key
                starting with the value. May not be used with excludeKey.
            excludeKey: When specified key(s) to exclude from the metadata. Matches any key
                starting with the value. May not be used with includeKey.
            expandSubWorkflows: When true, metadata for sub workflows will be fetched
                and inserted automatically in the metadata response.
            raise_for_status: Whether to check and raise for status based on the response.

        Raises:
            requests.exceptions.HTTPError: This will be raised when raise_for_status is True and Cromwell returns
                a response that satisfies 400 <= response.status_code < 600.

        Returns:
            HTTP response from Cromwell.
        """

        if excludeKey and includeKey:
            raise ValueError('includeKey and excludeKey may not be specified together!')

        params = {'expandSubWorkflows': json.dumps(expandSubWorkflows)}

        if isinstance(excludeKey, str):
            logger.info(f'Adding {excludeKey} to the request parameter list.')
            params['excludeKey'] = [excludeKey]
        elif isinstance(excludeKey, list) and len(excludeKey) >= 1:
            params['excludeKey'] = excludeKey

        if isinstance(includeKey, str):
            logger.info(f'Adding {includeKey} to the request parameter list.')
            params['includeKey'] = [includeKey]
        elif isinstance(includeKey, list) and len(includeKey) >= 1:
            params['includeKey'] = includeKey

        response = requests.get(
            url=auth.url + cls._metadata_endpoint.format(uuid=uuid),
            auth=auth.auth,
            headers=auth.header,
            params=params,
        )

        if raise_for_status:
            cls._check_and_raise_status(response)
        return response

    @classmethod
    def status(
        cls: 'CromwellAPI',
        uuid: str,
        auth: CromwellAuth,
        raise_for_status: bool = False,
    ) -> requests.Response:
        """Retrieves the current state for a workflow by UUID.

        Args:
            uuid: A Cromwell workflow UUID, which is the workflow identifier.
            auth: The authentication class holding headers or auth
                information to a Cromwell server.
            raise_for_status: Whether to check and raise for status based on the response.

        Raises:
            requests.exceptions.HTTPError: This will be raised when raise_for_status is True and Cromwell returns
                a response that satisfies 400 <= response.status_code < 600.

        Returns:
            HTTP response from Cromwell.
        """
        response = requests.get(
            url=auth.url + cls._status_endpoint.format(uuid=uuid),
            auth=auth.auth,
            headers=auth.header,
        )

        if raise_for_status:
            cls._check_and_raise_status(response)
        return response

    @classmethod
    def health(
        cls: 'CromwellAPI', auth: CromwellAuth, raise_for_status: bool = False
    ) -> requests.Response:
        """Return the current health status of any monitored subsystems of the Cromwell Server.

        Args:
            auth: authentication class holding headers or auth
                information to a Cromwell server.
            raise_for_status: Whether to check and raise for status based on the response.

        Raises:
            requests.exceptions.HTTPError: This will be raised when raise_for_status is True and Cromwell returns
                a response that satisfies 400 <= response.status_code < 600.

        Returns:
            HTTP response from Cromwell.
        """
        response = requests.get(
            url=auth.url + cls._health_endpoint, auth=auth.auth, headers=auth.header
        )

        if raise_for_status:
            cls._check_and_raise_status(response)
        return response

    @classmethod
    def submit(
        cls: 'CromwellAPI',
        auth: CromwellAuth,
        wdl_file: Union[str, io.BytesIO],
        inputs_files: Union[List[Union[str, io.BytesIO]], str, io.BytesIO] = None,
        options_file: Union[str, io.BytesIO] = None,
        dependencies: Union[str, List[str], io.BytesIO] = None,
        label_file: Union[str, io.BytesIO] = None,
        collection_name: str = None,
        on_hold: bool = False,
        validate_labels: bool = False,
        raise_for_status: bool = False,
    ) -> requests.Response:
        """ Submits a workflow to Cromwell.

        Args:
            auth: authentication class holding auth information to a
                Cromwell server.
            wdl_file: The workflow source file to submit for execution. Could be either the
                path to the file (str) or the file content in io.BytesIO.
            inputs_files: The input data in JSON
                format. Could be either the path to the file (str) or the file content in io.BytesIO. This could also
                be a list of unlimited input file paths/contents, each of them should have a type of
                Union[str, io.BytesIO].
            options_file: The Cromwell options file for workflows. Could be either
                the path to the file (str) or the file content in io.BytesIO.
            dependencies: Workflow dependency files. Could be the path to
                the zipped file (str) containing dependencies, a list of paths(List[str]) to all dependency files to be
                zipped or a zipped file in io.BytesIO.
            label_file: A collection of key/value pairs for workflow labels in JSON
                format, could be either the path to the JSON file (str) or the file content in io.BytesIO.
            collection_name: Collection in SAM that the workflow should belong to, if use CaaS.
            on_hold: Whether to submit the workflow in "On Hold" status.
            validate_labels: If True, validate cromwell labels.
            raise_for_status: Whether to check and raise for status based on the response.

        Raises:
            requests.exceptions.HTTPError: This will be raised when raise_for_status is True and Cromwell returns
                a response that satisfies 400 <= response.status_code < 600.

        Returns:
            HTTP response from Cromwell.
        """
        submission_manifest = utilities.prepare_workflow_manifest(
            wdl_file=wdl_file,
            inputs_files=inputs_files,
            options_file=options_file,
            dependencies=dependencies,
            label_file=label_file,
            collection_name=collection_name,
            on_hold=on_hold,
        )

        if auth.service_key_content:
            submission_manifest[
                'workflowOptions'
            ] = utilities.compose_oauth_options_for_jes_backend_cromwell(
                auth, submission_manifest.get('workflowOptions')
            )

        if validate_labels and label_file is not None:
            validate_cromwell_label(submission_manifest['labels'])

        response = requests.post(
            auth.url + cls._submit_endpoint,
            files=submission_manifest,
            auth=auth.auth,
            headers=auth.header,
        )

        if raise_for_status:
            cls._check_and_raise_status(response)
        return response

    @classmethod
    def wait(
        cls: 'CromwellAPI',
        workflow_ids: List[str],
        auth: CromwellAuth,
        timeout_minutes: int = 120,
        poll_interval_seconds: int = 30,
        verbose: bool = True,
    ) -> None:
        """Wait until cromwell returns successfully for each provided workflow

        Given a list of workflow ids, wait until cromwell returns successfully for each status, or
        one of the workflows fails or is aborted.

        Args:
            workflow_ids: A list of workflow ids to wait for terminal status.
            timeout_minutes: Maximum number of minutes to wait.
            auth: Authentication class holding headers
                or auth information to a Cromwell server.
            poll_interval_seconds: Number of seconds between checks for workflow
                completion.
            verbose: If True, report to stdout when all workflows succeed.
        """
        start = datetime.now()
        timeout = timedelta(minutes=int(timeout_minutes))

        while True:

            if datetime.now() - start > timeout:
                msg = f'Unfinished workflows after {timeout} minutes.'
                raise Exception(msg.format(timeout))

            all_succeeded = True

            if verbose:
                print('--- polling from cromwell ---')

            for uuid in workflow_ids:
                response = cls.status(uuid, auth)
                status = cls._parse_workflow_status(response)

                if verbose:
                    print(f'Workflow {uuid} returned status {status}')

                if status in _failed_statuses:
                    raise exceptions.WorkflowFailedError(
                        f'Workflow {uuid} returned status {status}'
                    )
                elif status != 'Succeeded':
                    all_succeeded = False

            if all_succeeded:
                print('All workflows succeeded!')
                return ''

            time.sleep(poll_interval_seconds)

    @classmethod
    def release_hold(
        cls: 'CromwellAPI',
        uuid: str,
        auth: CromwellAuth,
        raise_for_status: bool = False,
    ) -> requests.Response:
        """Request Cromwell to release the hold on a workflow.

        It will switch the status of a workflow from 'On Hold' to 'Submitted' so it can be picked for running. For
        a workflow that was not submitted with `workflowOnHold = true`, Cromwell will throw an error.

        Args:
            uuid: A Cromwell workflow UUID, which is the workflow identifier. The workflow is expected to have
                `On Hold` status.
            auth: The authentication class holding headers or auth
                information to a Cromwell server.
            raise_for_status: Whether to check and raise for status based on the response.

        Raises:
            requests.exceptions.HTTPError: This will be raised when raise_for_status is True and Cromwell returns
                a response that satisfies 400 <= response.status_code < 600.

        Returns:
            HTTP response from Cromwell.
        """
        response = requests.post(
            url=auth.url + cls._release_hold_endpoint.format(uuid=uuid),
            auth=auth.auth,
            headers=auth.header,
        )
        if raise_for_status:
            cls._check_and_raise_status(response)
        return response

    @classmethod
    def query(
        cls: 'CromwellAPI',
        query_dict: Dict[str, Union[str, List[str], Dict[str, str], bool]],
        auth: CromwellAuth,
        raise_for_status: bool = False,
    ) -> requests.Response:
        """Query for workflows.

        TODO: Given that Cromwell-as-a-Service blocks a set of features that are available in Cromwell, e.g. 'labelor',
        for security concerns, the first iteration of this API doesn't come up with the advanced query keys of the
        Cromwell except a set of necessary ones. However, we need to implement this for completeness and keep an eye
        on the compatibility between CaaS and Cromwell.

        All of the query keys will be used in an OR manner, except the keys within `labels`, which are defined in
        an AND relation. For instance, [{'status': 'Succeeded'}, {'status': 'Failed'}] will give you all of the
        workflows that in either `Succeeded` or `Failed` statuses.

        Args:
            query_dict: A dictionary representing the query key-value paris. The keys should be accepted by the
                Cromwell or they will get ignored. The values could be str, list or dict.
            auth: The authentication class holding headers or auth
                information to a Cromwell server.
            raise_for_status: Whether to check and raise for status based on the response.

        Raises:
            requests.exceptions.HTTPError: This will be raised when raise_for_status is True and Cromwell returns
                a response that satisfies 400 <= response.status_code < 600.

        Returns:
            HTTP response from Cromwell.
        """
        if (
            'additionalQueryResultFields' in query_dict.keys()
            or 'includeSubworkflows' in query_dict.keys()
        ):
            logging.warning(
                'Note: additionalQueryResultFields, includeSubworkflows may not scale due to the '
                'following issues with Cromwell: https://github.com/broadinstitute/cromwell/issues/3115 '
                'and https://github.com/broadinstitute/cromwell/issues/3873'
            )

        query_params = cls._compose_query_params(query_dict)

        response = requests.post(
            url=auth.url + cls._query_endpoint,
            json=query_params,
            auth=auth.auth,
            headers=auth.header,
        )
        if raise_for_status:
            cls._check_and_raise_status(response)
        return response

    @classmethod
    def _compose_query_params(
        cls: 'CromwellAPI',
        query_dict: Dict[str, Union[str, List[str], Dict[str, str], bool]],
    ) -> List[Dict[str, str]]:
        """Helper function to compose the query params that could be accepted by Cromwell.

        This function will parse and compose the query params for Cromwell's /query endpoint from an user's input
        query dictionary. It also provides very basic inputs validation so users don't have to wait for the error
        response from Cromwell for a long time.

        The query keys should be one of the following strings in the `cls._cromwell_query_keys` set, otherwise
        they will be ignore by this function.

        In general, this method is expecting the input query dictionary follows a basic
        structure like below:

        ```
        query_dict = {
            'label': {
                'cromwell-workflow-id': 'cromwell-xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
            },
            'status': ['Running', 'Succeeded'],
            'id': 'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx',
            'additionalQueryResultFields': 'labels',
            'submission': '2018-01-01T00:01:01.410150Z',
            'start': '2018-01-01T01:01:01.410150Z',
            'end': '2018-01-01T02:01:01.410150Z',
            'name': ['WorkflowName1', 'WorkflowName2'],
            'additionalQueryResultFields': ['labels', 'parentWorkflowId'],
            'includeSubworkflows': True
        }
        ```

        which will be converted to the following query parameters:

        ```
        query_params = [
            {'label': 'cromwell-workflow-id:cromwell-xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'},
            {'status': 'Running'},
            {'status': 'Succeeded'},
            {'id': 'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'},
            {'additionalQueryResultFields': 'labels'},
            {'submission': '2018-01-01T00:01:01.410150Z'},
            {'start': '2018-01-01T01:01:01.410150Z'},
            {'end': '2018-01-01T02:01:01.410150Z'},
            {'name': 'WorkflowName1'},
            {'name': 'WorkflowName2'},
            {'additionalQueryResultFields': 'labels'},
            {'additionalQueryResultFields': 'parentWorkflowId'},
            {'includeSubworkflows': 'true'}
        ]
        ```

        Args:
            query_dict: A dictionary representing the query key-value paris. The keys should be accepted by the
                Cromwell or they will get ignored. The values could be str, list or dict.

        Raises:
            TypeError: If the input query_dict is not a dictionary.
            ValueError: If a list of values are assigned to a query key that belongs to _cromwell_exclusive_query_keys.

        Returns:
            query_params: A composed list of query objects.
        """
        if not isinstance(query_dict, dict):
            raise TypeError('A valid dictionary with query keys is required!')

        query_params = []
        for k, v in query_dict.items():
            if k in _cromwell_query_keys:
                if k == 'label' and isinstance(v, dict):
                    query_params.extend(
                        [
                            {'label': label_key + ':' + label_value}
                            for label_key, label_value in v.items()
                        ]
                    )
                elif isinstance(v, list):
                    if k in _cromwell_exclusive_query_keys:
                        raise ValueError(
                            '{} cannot be specified multiple times!'.format(k)
                        )
                    query_params.extend(
                        [
                            {k: json.dumps(val)}
                            if not isinstance(val, str)
                            else {k: val}
                            for val in set(v)
                        ]
                    )
                else:
                    query_params.append(
                        {k: json.dumps(v)} if not isinstance(v, str) else {k: v}
                    )
            else:
                logger.info(
                    '{} is not an allowed query key in Cromwell, will be ignored in this query.'.format(
                        k
                    )
                )
        return query_params

    @staticmethod
    def _parse_workflow_status(response: requests.Response) -> str:
        """Helper function to parse a status response.

        Args:
            response: A status response object from Cromwell.

        Raises:
            WorkflowUnknownError: This will be raised when Cromwell returns a status code != 200.

        Returns:
            String representing status response.
        """
        if response.status_code != 200:
            raise exceptions.WorkflowUnknownError(
                'Status could not be determined, endpoint returned {0}'.format(
                    response.status_code
                )
            )
        else:
            return response.json()['status']

    @staticmethod
    def _check_and_raise_status(response: requests.Response) -> None:
        """Helper function to check the status of a response and raise a friendly message if there are errors.

        This functions is using the `response.ok` which wraps the `raise_for_status()`, by doing this, we can
        produce the actual error messages from the Cromwell, instead of shadowing them with `raise_for_status()`.

        Args:
            response: A status response object from Cromwell.

        Raises:
            requests.exceptions.HTTPError: This will be raised when Cromwell returns a response that satisfies
                400 <= response.status_code < 600.
        """
        if not response.ok:
            raise requests.exceptions.HTTPError(
                'Error Code {0}: {1}'.format(response.status_code, response.text)
            )



import io
import json
import os
import requests
import requests_mock
import six
import tempfile
import unittest
import zipfile


six.add_move(six.MovedModule('mock', 'mock', 'unittest.mock'))
from six.moves import mock  # noqa

from cromwell_tools import utilities as utils  # noqa
from cromwell_tools.cromwell_auth import CromwellAuth  # noqa


class TestUtilities(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Change to test directory, as tests may have been invoked from another dir
        dir_ = os.path.abspath(os.path.dirname(__file__))
        os.chdir(dir_)

    @staticmethod
    def _load_test_data_as_BytesIO(path):
        with open(path, 'rb') as fp:
            stream = io.BytesIO(fp.read())
        return stream

    def setUp(self):
        self.invalid_labels = {
            "0-label-key-1": "0-label-value-1",
            "the-maximum-allowed-character-length-for-label-pairs-is-sixty-three": "cromwell-please-dont-validate-these-labels",
            "": "not a great label key",
            "Comment": "This-is-a-test-label",
        }
        self.valid_labels = {
            "label-key-1": "label-value-1",
            "label-key-2": "label-value-2",
            "only-key": "",
            "fc-id": "0123-abcd-4567-efgh",
            "comment": "this-is-a-test-label",
        }
        self.wdl_file_path = 'data/test_workflow.wdl'
        self.wdl_file_BytesIO = self._load_test_data_as_BytesIO(self.wdl_file_path)

        self.inputs_file_path = 'data/test_inputs1.json'
        self.inputs_file_BytesIO = self._load_test_data_as_BytesIO(
            self.inputs_file_path
        )

        self.inputs_file_path_list = [
            'data/test_inputs1.json',
            'data/test_inputs2.json',
        ]
        self.inputs_file_BytesIO_list = [
            self._load_test_data_as_BytesIO(f) for f in self.inputs_file_path_list
        ]

        self.options_file_path = 'data/test_options.json'
        self.options_file_BytesIO = self._load_test_data_as_BytesIO(
            self.options_file_path
        )

        self.label_file_path = 'data/test_labels.json'
        self.label_file_BytesIO = self._load_test_data_as_BytesIO(self.label_file_path)

        self.deps_zip_file_path = 'data/test_deps.zip'
        self.deps_files_paths_list = ['data/test_task.wdl']
        self.deps_zip_file_BytesIO = self._load_test_data_as_BytesIO(
            self.deps_zip_file_path
        )

    @requests_mock.mock()
    def test_download_http_raises_error_on_bad_status_code(self, mock_request):
        def _request_callback(request, context):
            context.status_code = 404
            return 'Not found'

        fake_url = 'https://fake_url'
        mock_request.get(fake_url, json=_request_callback)

        with self.assertRaises(requests.HTTPError):
            utils.download_http(fake_url)

    @requests_mock.mock()
    def test_download_http_no_error_on_200(self, mock_request):
        def _request_callback(request, context):
            context.status_code = 200
            return 'foo'

        fake_url = 'https://fake_url'

        mock_request.get(fake_url, json=_request_callback)
        try:
            utils.download_http(fake_url)
        except requests.HTTPError:
            self.fail('Raised HTTPError for status 200')

    @requests_mock.mock()
    def test_download_http_no_error_on_301(self, mock_request):
        def _request_callback(request, context):
            context.status_code = 301
            return 'foo'

        fake_url = 'https://fake_url'
        mock_request.get(fake_url, json=_request_callback)
        try:
            utils.download_http(fake_url)
        except requests.HTTPError:
            self.fail('Raised HTTPError for status 301')

    def test_download_to_map(self):
        """Test download_to_map with local files to ensure it builds the map of paths to contents correctly"""
        urls = ['data/a.txt', 'data/b.txt']
        urls_to_content = utils.download_to_map(urls)
        self.assertIn('data/a.txt', urls_to_content)
        self.assertIn('data/b.txt', urls_to_content)
        self.assertEqual(urls_to_content['data/a.txt'], b'aaa\n')
        self.assertEqual(urls_to_content['data/b.txt'], b'bbb\n')

    def test_make_zip_in_memory(self):
        """Test make_zip_in_memory produces an in-memory zip file with the expected contents"""

        # Encoding the values below gives more realistic inputs for this test.
        # It gives us str instances in Python 2 and bytes instances in Python 3,
        # which is what is actually received respectively in these two versions by
        # make_zip_in_memory.
        urls_to_content = {
            'data/a.txt': 'aaa\n'.encode('utf-8'),
            'data/b.txt': 'bbb\n'.encode('utf-8'),
        }
        bytes_buf = utils.make_zip_in_memory(urls_to_content)
        with zipfile.ZipFile(bytes_buf, 'r') as zf:
            with zf.open('a.txt') as f1:
                f1_contents = f1.read()
            with zf.open('b.txt') as f2:
                f2_contents = f2.read()
        self.assertEqual(f1_contents, b'aaa\n')
        self.assertEqual(f2_contents, b'bbb\n')

    def test_validate_cromwell_label_on_invalid_labels_object(self):
        self.assertRaises(
            ValueError, utils.validate_cromwell_label, self.invalid_labels
        )

    def test_validate_cromwell_label_on_invalid_labels_str_object(self):
        self.assertRaises(
            ValueError, utils.validate_cromwell_label, json.dumps(self.invalid_labels)
        )

    def test_validate_cromwell_label_on_invalid_labels_bytes_object(self):
        self.assertRaises(
            ValueError,
            utils.validate_cromwell_label,
            json.dumps(self.invalid_labels).encode('utf-8'),
        )

    def test_validate_cromwell_label_on_valid_labels_object(self):
        self.assertIsNone(utils.validate_cromwell_label(self.valid_labels))

    def test_validate_cromwell_label_on_valid_labels_str_object(self):
        self.assertIsNone(utils.validate_cromwell_label(json.dumps(self.valid_labels)))

    def test_validate_cromwell_label_on_valid_labels_bytes_object(self):
        self.assertIsNone(
            utils.validate_cromwell_label(json.dumps(self.valid_labels).encode('utf-8'))
        )

    def test_localize_file(self):
        temporary_directory = tempfile.mkdtemp()
        # test that we can localize both local and https files. Use this file as a convenient target
        targets = (
            'https://raw.githubusercontent.com/broadinstitute/cromwell-tools/'
            'v0.5.0/cromwell_tools/tests/test_cromwell_tools.py',
            __file__,
        )
        for target in targets:
            utils._localize_file(target, temporary_directory)
            localized_file = os.path.join(temporary_directory, os.path.basename(target))

            # verify the file was downloaded and that it contains some content we expect
            assert os.path.isfile(localized_file)
            with open(localized_file, 'r') as f:
                assert 'cromwell_tools' in f.read()
            os.remove(localized_file)

    def test_prepare_workflow_manifest_works_for_wdl_file_with_filepath(self):
        manifest = utils.prepare_workflow_manifest(wdl_file=self.wdl_file_path)
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowSource'].getvalue()
            == expected_manifest['workflowSource'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_wdl_file_with_BytesIO(self):
        manifest = utils.prepare_workflow_manifest(wdl_file=self.wdl_file_BytesIO)
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowSource'].getvalue()
            == expected_manifest['workflowSource'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_one_inputs_file_with_filepath(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, inputs_files=self.inputs_file_path
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowInputs': self.inputs_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowInputs'].getvalue()
            == expected_manifest['workflowInputs'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_one_inputs_file_with_BytesIO(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, inputs_files=self.inputs_file_BytesIO
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowInputs': self.inputs_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowInputs'].getvalue()
            == expected_manifest['workflowInputs'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_multiple_inputs_files_with_filepath(
        self
    ):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, inputs_files=self.inputs_file_path_list
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowInputs': self.inputs_file_BytesIO_list[0],
            'workflowInputs_2': self.inputs_file_BytesIO_list[1],
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowInputs'].getvalue()
            == expected_manifest['workflowInputs'].getvalue()
        )
        assert (
            manifest['workflowInputs_2'].getvalue()
            == expected_manifest['workflowInputs_2'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_multiple_inputs_files_with_BytesIO(
        self
    ):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, inputs_files=self.inputs_file_BytesIO_list
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowInputs': self.inputs_file_BytesIO_list[0],
            'workflowInputs_2': self.inputs_file_BytesIO_list[1],
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowInputs'].getvalue()
            == expected_manifest['workflowInputs'].getvalue()
        )
        assert (
            manifest['workflowInputs_2'].getvalue()
            == expected_manifest['workflowInputs_2'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_dependencies_file_with_filepath(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, dependencies=self.deps_zip_file_path
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowDependencies': self.deps_zip_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowDependencies'].getvalue()
            == expected_manifest['workflowDependencies'].getvalue()
        )

    def test_prepare_workflow_manifest_raises_an_error_for_dependencies_file_with_filepath_not_pointing_to_zip(
        self
    ):
        with self.assertRaises(ValueError):
            utils.prepare_workflow_manifest(
                wdl_file=self.wdl_file_path, dependencies="data/fake_test_deps.wdl"
            )

    def test_prepare_workflow_manifest_works_for_dependencies_file_with_filepath_in_a_list(
        self
    ):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, dependencies=[self.deps_zip_file_path]
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowDependencies': self.deps_zip_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowDependencies'].getvalue()
            == expected_manifest['workflowDependencies'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_dependencies_file_with_list_of_files(
        self
    ):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, dependencies=self.deps_files_paths_list
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowDependencies': self.deps_zip_file_BytesIO,
            'workflowOnHold': 'false',
        }
        with zipfile.ZipFile(manifest['workflowDependencies'], 'r') as zf:
            with zf.open('test_task.wdl') as dep_file:
                contents = io.BytesIO(dep_file.read())

        with zipfile.ZipFile(expected_manifest['workflowDependencies'], 'r') as zf:
            with zf.open('test_task.wdl') as dep_file:
                expected_contents = io.BytesIO(dep_file.read())

        # we have to unzip both of them to get the proper comparison during the test
        assert contents.getvalue() == expected_contents.getvalue()

    def test_prepare_workflow_manifest_works_for_dependencies_file_with_BytesIO(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, dependencies=self.deps_zip_file_BytesIO
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowDependencies': self.deps_zip_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowDependencies'].getvalue()
            == expected_manifest['workflowDependencies'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_collection_name(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, collection_name='test_collection'
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'collectionName': 'test_collection',
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowSource'].getvalue()
            == expected_manifest['workflowSource'].getvalue()
        )
        assert manifest['collectionName'] == expected_manifest['collectionName']

    def test_prepare_workflow_manifest_works_for_on_hold(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, on_hold=True
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowOnHold': 'true',
        }
        assert (
            manifest['workflowSource'].getvalue()
            == expected_manifest['workflowSource'].getvalue()
        )
        assert manifest['workflowOnHold'] == expected_manifest['workflowOnHold']

    def test_prepare_workflow_manifest_works_for_label_file_with_filepath(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, label_file=self.label_file_path
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'labels': self.label_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert manifest['labels'].getvalue() == expected_manifest['labels'].getvalue()

    def test_prepare_workflow_manifest_works_for_label_file_with_BytesIO(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, label_file=self.label_file_BytesIO
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'labels': self.label_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert manifest['labels'].getvalue() == expected_manifest['labels'].getvalue()

    def test_prepare_workflow_manifest_works_for_options_file_with_filepath(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, options_file=self.options_file_path
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowOptions': self.options_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowOptions'].getvalue()
            == expected_manifest['workflowOptions'].getvalue()
        )

    def test_prepare_workflow_manifest_works_for_options_file_with_BytesIO(self):
        manifest = utils.prepare_workflow_manifest(
            wdl_file=self.wdl_file_path, options_file=self.options_file_BytesIO
        )
        expected_manifest = {
            'workflowSource': self.wdl_file_BytesIO,
            'workflowOptions': self.options_file_BytesIO,
            'workflowOnHold': 'false',
        }
        assert (
            manifest['workflowOptions'].getvalue()
            == expected_manifest['workflowOptions'].getvalue()
        )

    def test_compose_oauth_options_for_jes_backend_cromwell_add_required_fields_to_workflow_options(
        self
    ):
        test_url = 'https://fake_url'
        test_service_account_key = 'data/fake_account_key.json'
        with open(test_service_account_key, 'r') as f:
            test_service_account_key_content = json.load(f)

        test_auth = CromwellAuth(
            url=test_url,
            header={"Authorization": "bearer fake_token"},
            auth=None,
            service_key_content=test_service_account_key_content,
        )

        result_options = utils.compose_oauth_options_for_jes_backend_cromwell(
            test_auth, self.options_file_BytesIO
        )
        result_options_in_dict = json.loads(result_options.getvalue())

        assert (
            result_options_in_dict['read_from_cache']
            == json.loads(self.options_file_BytesIO.getvalue())['read_from_cache']
        )
        assert (
            result_options_in_dict['google_project']
            == test_service_account_key_content['project_id']
        )
        assert (
            result_options_in_dict['google_compute_service_account']
            == test_service_account_key_content['client_email']
        )
        assert result_options_in_dict['user_service_account_json'] == json.dumps(
            test_service_account_key_content
        )

    def test_if_compose_oauth_options_for_jes_backend_cromwell_can_deal_with_null_workflow_options(
        self
    ):
        test_url = 'https://fake_url'
        test_service_account_key = 'data/fake_account_key.json'
        with open(test_service_account_key, 'r') as f:
            test_service_account_key_content = json.load(f)

        test_auth = CromwellAuth(
            url=test_url,
            header={"Authorization": "bearer fake_token"},
            auth=None,
            service_key_content=test_service_account_key_content,
        )

        result_options = utils.compose_oauth_options_for_jes_backend_cromwell(test_auth)
        result_options_in_dict = json.loads(result_options.getvalue())

        assert (
            result_options_in_dict['google_project']
            == test_service_account_key_content['project_id']
        )
        assert (
            result_options_in_dict['google_compute_service_account']
            == test_service_account_key_content['client_email']
        )
        assert result_options_in_dict['user_service_account_json'] == json.dumps(
            test_service_account_key_content
        )



#!/usr/bin/env python
import io
import json
import os
import requests
import requests_mock
import six
import tempfile
import unittest


six.add_move(six.MovedModule('mock', 'mock', 'unittest.mock'))
from six.moves import mock  # noqa

from cromwell_tools.cromwell_api import CromwellAPI  # noqa
from cromwell_tools.cromwell_auth import CromwellAuth  # noqa
from cromwell_tools import utilities as utils  # noqa


class TestAPI(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Change to test directory, as tests may have been invoked from another dir
        dir_ = os.path.abspath(os.path.dirname(__file__))
        os.chdir(dir_)

    def setUp(self):
        self.wdl_file = io.BytesIO(b"wdl_file_content")
        self.zip_file = io.BytesIO(b"zip_file_content")
        self.inputs_file = io.BytesIO(b"inputs_file_content")
        self.options_file = io.BytesIO(b"options_file_content")
        self.label = io.BytesIO(b'{"test-label-key": "test-label-value"}')
        self.auth_options = self.set_up_auth()

    @mock.patch(
        'cromwell_tools.cromwell_auth.CromwellAuth.from_service_account_key_file'
    )
    def set_up_auth(self, mock_header):
        # set up authentication options for the tests
        temp_dir = tempfile.mkdtemp()
        secrets_file = temp_dir + 'fake_secrets.json'
        service_account_key = os.path.join(temp_dir, 'fake_key.json')
        username = "fake_user"
        password = "fake_password"
        url = "https://fake_url"
        auth = {"url": url, "username": username, "password": password}
        with open(secrets_file, 'w') as f:
            json.dump(auth, f)
        mock_header.return_value = CromwellAuth(
            url=url, header={"Authorization": "bearer fake_token"}, auth=None
        )

        auth_options = (
            CromwellAuth.harmonize_credentials(**auth),  # HTTPBasicAuth
            CromwellAuth.harmonize_credentials(
                **{"secrets_file": secrets_file}
            ),  # Secret file
            CromwellAuth.harmonize_credentials(
                **{"service_account_key": service_account_key, "url": url}
            ),  # OAuth
            CromwellAuth.harmonize_credentials(url=url),  # No Auth
        )
        return auth_options

    def _submit_workflows(self, cromwell_auth, mock_request, _request_callback):
        mock_request.post(
            cromwell_auth.url + '/api/workflows/v1', json=_request_callback
        )
        return CromwellAPI.submit(
            auth=cromwell_auth,
            wdl_file=self.wdl_file,
            inputs_files=self.inputs_file,
            options_file=self.options_file,
            dependencies=self.zip_file,
            label_file=self.label,
        )

    @requests_mock.mock()
    def test_submit_workflow(self, mock_request):
        def _request_callback(request, context):
            context.status_code = 200
            context.headers['test'] = 'header'
            return {'request': {'body': "content"}}

        for cromwell_auth in self.auth_options:
            result = self._submit_workflows(
                cromwell_auth, mock_request, _request_callback
            )
            self.assertEqual(result.status_code, 200)
            self.assertEqual(result.headers.get('test'), 'header')

    @requests_mock.mock()
    def test_submit_workflow_handlers_error_response(self, mock_request):
        def _request_callback(request, context):
            context.status_code = 500
            context.headers['test'] = 'header'
            return {'status': 'error', 'message': 'Internal Server Error'}

        # Check request actions
        for cromwell_auth in self.auth_options:
            with self.assertRaises(requests.HTTPError):
                self._submit_workflows(
                    cromwell_auth, mock_request, _request_callback
                ).raise_for_status()

    @requests_mock.mock()
    def test_query_workflows_returns_200(self, mock_request):
        query_dict = {
            'status': ['Running', 'Failed'],
            'label': {'label_key1': 'label_value1', 'label_key2': 'label_value2'},
        }

        def _request_callback(request, context):
            context.status_code = 200
            context.headers['test'] = 'header'
            return {
                'results': [
                    {
                        'name': 'workflow1',
                        'submission': 'submission1',
                        'id': 'id1',
                        'status': 'Failed',
                        'start': 'start1',
                        'end': 'end1',
                    },
                    {
                        'name': 'workflow2',
                        'submission': 'submission2',
                        'id': 'id2',
                        'status': 'Running',
                        'start': 'start2',
                        'end': 'end2',
                    },
                ],
                'totalResultsCount': 2,
            }

        for cromwell_auth in self.auth_options:
            mock_request.post(
                '{}/api/workflows/v1/query'.format(cromwell_auth.url),
                json=_request_callback,
            )
            result = CromwellAPI.query(query_dict, cromwell_auth)
            self.assertEqual(result.status_code, 200)
            self.assertEqual(result.json()['totalResultsCount'], 2)

    def test_compose_query_params_can_compose_simple_query_dicts(self):
        query_dict = {
            'status': 'Running',
            'start': '2018-01-01T00:00:00.000Z',
            'end': '2018-01-01T12:00:00.000Z',
            'label': {'Comment': 'test'},
            'page': 1,
            'pageSize': 10,
        }

        expect_params = [
            {'status': 'Running'},
            {'start': '2018-01-01T00:00:00.000Z'},
            {'end': '2018-01-01T12:00:00.000Z'},
            {'label': 'Comment:test'},
            {'page': '1'},
            {'pageSize': '10'},
        ]

        six.assertCountEqual(
            self, CromwellAPI._compose_query_params(query_dict), expect_params
        )

    def test_compose_query_params_can_compose_nested_query_dicts(self):
        query_dict = {
            'status': ['Running', 'Failed', 'Submitted'],
            'start': '2018-01-01T00:00:00.000Z',
            'end': '2018-01-01T12:00:00.000Z',
            'label': {'Comment1': 'test1', 'Comment2': 'test2', 'Comment3': 'test3'},
        }

        expect_params = [
            {'status': 'Running'},
            {'status': 'Failed'},
            {'status': 'Submitted'},
            {'start': '2018-01-01T00:00:00.000Z'},
            {'end': '2018-01-01T12:00:00.000Z'},
            {'label': 'Comment1:test1'},
            {'label': 'Comment2:test2'},
            {'label': 'Comment3:test3'},
        ]
        six.assertCountEqual(
            self, CromwellAPI._compose_query_params(query_dict), expect_params
        )

    def test_compose_query_params_can_convert_bools_within_query_dicts(self):
        query_dict = {
            'status': ['Running', 'Failed', 'Submitted'],
            'start': '2018-01-01T00:00:00.000Z',
            'end': '2018-01-01T12:00:00.000Z',
            'label': {'Comment1': 'test1', 'Comment2': 'test2', 'Comment3': 'test3'},
            'includeSubworkflows': True,
        }

        expect_params = [
            {'status': 'Running'},
            {'status': 'Failed'},
            {'status': 'Submitted'},
            {'start': '2018-01-01T00:00:00.000Z'},
            {'end': '2018-01-01T12:00:00.000Z'},
            {'label': 'Comment1:test1'},
            {'label': 'Comment2:test2'},
            {'label': 'Comment3:test3'},
            {'includeSubworkflows': 'true'},
        ]
        six.assertCountEqual(
            self, CromwellAPI._compose_query_params(query_dict), expect_params
        )

    def test_compose_query_params_raises_error_for_invalid_query_dict_that_has_multiple_values_for_exclusive_keys(
        self
    ):
        query_dict = {
            'status': ['Running', 'Failed', 'Submitted'],
            'start': ['2018-01-01T00:00:00.000Z', '2018-01-02T00:00:00.000Z'],
            'end': '2018-01-01T12:00:00.000Z',
            'label': {'Comment1': 'test1', 'Comment2': 'test2', 'Comment3': 'test3'},
        }

        with self.assertRaises(ValueError):
            CromwellAPI._compose_query_params(query_dict)

    @requests_mock.mock()
    def test_release_onhold_returns_200(self, mock_request):
        workflow_id = '12345abcde'

        def _request_callback(request, context):
            context.status_code = 200
            context.headers['test'] = 'header'
            return {'id': request.url.split('/')[-2], 'status': 'Submitted'}

        for cromwell_auth in self.auth_options:
            mock_request.post(
                '{0}/api/workflows/v1/{1}/releaseHold'.format(
                    cromwell_auth.url, workflow_id
                ),
                json=_request_callback,
            )
            result = CromwellAPI.release_hold(workflow_id, cromwell_auth)
            self.assertEqual(result.status_code, 200)
            self.assertEqual(result.json()['id'], workflow_id)
            self.assertEqual(result.json()['status'], 'Submitted')

    @requests_mock.mock()
    def test_release_workflow_that_is_not_on_hold_returns_error(self, mock_request):
        workflow_id = 'test'

        def _request_callback(request, context):
            context.status_code = 403
            context.headers['test'] = 'header'
            return {
                'status': 'error',
                'message': 'Couldn\'t change status of workflow {} to \'Submitted\' because the workflow is not in '
                '\'On Hold\' state'.format(request.url.split('/')[-2]),
            }

        for cromwell_auth in self.auth_options:
            mock_request.post(
                '{0}/api/workflows/v1/{1}/releaseHold'.format(
                    cromwell_auth.url, workflow_id
                ),
                json=_request_callback,
            )
            with self.assertRaises(requests.exceptions.HTTPError):
                CromwellAPI.release_hold(workflow_id, cromwell_auth).raise_for_status()

    @requests_mock.mock()
    def test_metadata_returns_200(self, mock_request):
        workflow_id = '12345abcde'
        test_include_key = 'workflow'

        def _request_callback(request, context):
            context.status_code = 200
            context.headers['test'] = 'header'
            return {'id': '12345abcde', 'actualWorkflowLanguageVersion': 'draft-2'}

        for cromwell_auth in self.auth_options:
            mock_request.get(
                '{0}/api/workflows/v1/{1}/metadata?expandSubWorkflows=false&includeKey={2}'.format(
                    cromwell_auth.url, workflow_id, test_include_key
                ),
                json=_request_callback,
            )
            result = CromwellAPI.metadata(
                workflow_id, cromwell_auth, includeKey=test_include_key
            )
            self.assertEqual(result.status_code, 200)
            self.assertEqual(result.json()['id'], workflow_id)

    @requests_mock.mock()
    def test_health_returns_200(self, mock_request):
        expected = {
            "DockerHub": {"ok": "true"},
            "Engine Database": {"ok": "true"},
            "PAPI": {"ok": "true"},
            "GCS": {"ok": "true"},
        }

        def _request_callback(request, context):
            context.status_code = 200
            context.headers['test'] = 'header'
            return expected

        for cromwell_auth in self.auth_options:
            mock_request.get(
                '{0}/engine/v1/status'.format(cromwell_auth.url), json=_request_callback
            )
            result = CromwellAPI.health(cromwell_auth)
            self.assertEqual(result.status_code, 200)
            self.assertEqual(result.json(), expected)

    @requests_mock.mock()
    def test_abort(self, mock_request):
        workflow_id = "01234"
        expected = {"id": workflow_id, "status": "Aborting"}

        def _request_callback(request, context):
            context.status_code = 200
            context.headers['test'] = 'header'
            return expected

        for cromwell_auth in self.auth_options:
            mock_request.post(
                cromwell_auth.url + '/api/workflows/v1/{}/abort'.format(workflow_id),
                json=_request_callback,
            )
            result = CromwellAPI.abort(workflow_id, cromwell_auth)
            self.assertEqual(result.json(), expected)

    @requests_mock.mock()
    def test_status(self, mock_request):
        def _request_callback_status(request, context):
            context.status_code = 200
            context.headers['test'] = 'header'
            return {'status': 'Succeeded'}

        workflow_id = "01234"
        for cromwell_auth in self.auth_options:
            mock_request.get(
                cromwell_auth.url + '/api/workflows/v1/{}/status'.format(workflow_id),
                json=_request_callback_status,
            )
            result = CromwellAPI.status(workflow_id, cromwell_auth)
            self.assertEqual(result.json()['status'], 'Succeeded')


if __name__ == '__main__':
    unittest.main()



import argparse
import requests
from cromwell_tools.cromwell_api import CromwellAPI
from cromwell_tools.cromwell_auth import CromwellAuth
from cromwell_tools import __version__


def parser(arguments=None):
    # TODO: dynamically walk through the commands and automatcally create parsers here

    main_parser = argparse.ArgumentParser()

    # Check the installed version of Cromwell-tools
    main_parser.add_argument(
        '-V', '--version', action='version', version=f'%(prog)s {__version__}'
    )

    subparsers = main_parser.add_subparsers(help='sub-command help', dest='command')

    # sub-commands of cromwell-tools
    submit = subparsers.add_parser(
        'submit', help='submit help', description='Submit a WDL workflow on Cromwell.'
    )
    wait = subparsers.add_parser(
        'wait',
        help='wait help',
        description='Wait for one or more running workflow to finish.',
    )
    status = subparsers.add_parser(
        'status',
        help='status help',
        description='Get the status of one or more workflows.',
    )
    abort = subparsers.add_parser(
        'abort',
        help='abort help',
        description='Request Cromwell to abort a running workflow by UUID.',
    )
    release_hold = subparsers.add_parser(
        'release_hold',
        help='release_hold help',
        description='Request Cromwell to release the hold on a workflow.',
    )
    metadata = subparsers.add_parser(
        'metadata',
        help='metadata help',
        description='Retrieve the workflow and call-level metadata for a specified workflow by UUID.',
    )
    query = subparsers.add_parser(
        'query',
        help='query help',
        description='[NOT IMPLEMENTED IN CLI] Query for workflows.',
    )
    health = subparsers.add_parser(
        'health',
        help='health help',
        description='Check that cromwell is running and that provided authentication is valid.',
    )

    # cromwell url and authentication arguments apply to all sub-commands
    cromwell_sub_commands = (
        submit,
        wait,
        status,
        abort,
        release_hold,
        metadata,
        query,
        health,
    )
    auth_args = {
        'url': 'The URL to the Cromwell server. e.g. "https://cromwell.server.org/"',
        'username': 'Cromwell username for HTTPBasicAuth.',
        'password': 'Cromwell password for HTTPBasicAuth.',
        'secrets_file': 'Path to the JSON file containing username, password, and url fields.',
        'service_account_key': 'Path to the JSON key file for authenticating with CaaS.',
    }

    def add_auth_args(subcommand_parser):
        for arg_dest, help_text in auth_args.items():
            subcommand_parser.add_argument(
                '--{arg}'.format(arg=arg_dest.replace('_', '-')),
                dest=arg_dest,
                default=None,
                type=str,
                help=help_text,
            )

    # TODO: this should be a group which is called authentication
    for p in cromwell_sub_commands:
        add_auth_args(p)

    # submit arguments
    submit.add_argument(
        '-w',
        '--wdl-file',
        dest='wdl_file',
        type=str,
        required=True,
        help='Path to the workflow source file to submit for execution.',
    )
    submit.add_argument(
        '-i',
        '--inputs-files',
        dest='inputs_files',
        nargs='+',
        type=str,
        required=True,
        help='Path(s) to the input file(s) containing input data in JSON format, separated by space.',
    )
    submit.add_argument(
        '-d',
        '--deps-file',
        dest='dependencies',
        nargs='+',
        type=str,
        help='Path to the Zip file containing dependencies, or a list of raw dependency files to '
        'be zipped together separated by space.',
    )
    submit.add_argument(
        '-o',
        '--options-file',
        dest='options_file',
        type=str,
        help='Path to the Cromwell configs JSON file.',
    )
    # TODO: add a mutually exclusive group to make it easy to add labels for users
    submit.add_argument(
        '-l',
        '--label-file',
        dest='label_file',
        type=str,
        default=None,
        help='Path to the JSON file containing a collection of key/value pairs for workflow labels.',
    )
    submit.add_argument(
        '-c',
        '--collection-name',
        dest='collection_name',
        type=str,
        default=None,
        help='Collection in SAM that the workflow should belong to, if use CaaS.',
    )
    submit.add_argument(
        '--on-hold',
        dest='on_hold',
        type=bool,
        default=False,
        help='Whether to submit the workflow in "On Hold" status.',
    )
    submit.add_argument(
        '--validate-labels',
        dest='validate_labels',
        type=bool,
        default=False,
        help='Whether to validate cromwell labels.',
    )

    # wait arguments
    wait.add_argument('workflow_ids', nargs='+')
    wait.add_argument(
        '--timeout-minutes',
        dest='timeout_minutes',
        type=int,
        default=120,
        help='number of minutes to wait before timeout.',
    )
    wait.add_argument(
        '--poll-interval-seconds',
        dest='poll_interval_seconds',
        type=int,
        default=30,
        help='seconds between polling cromwell for workflow status.',
    )
    wait.add_argument(
        '--silent',
        dest='verbose',
        action='store_false',
        help='whether to silently print verbose workflow information while polling cromwell.',
    )

    # status arguments
    status.add_argument(
        '--uuid',
        required=True,
        help='A Cromwell workflow UUID, which is the workflow identifier.',
    )

    # abort arguments
    abort.add_argument(
        '--uuid',
        required=True,
        help='A Cromwell workflow UUID, which is the workflow identifier.',
    )

    # release_hold arguments
    release_hold.add_argument(
        '--uuid',
        required=True,
        help='A Cromwell workflow UUID, which is the workflow identifier.',
    )

    # metadata arguments
    metadata.add_argument(
        '--uuid',
        required=True,
        help='A Cromwell workflow UUID, which is the workflow identifier.',
    )
    # TODO: add a mutually exclusive group to make it fail early
    metadata.add_argument(
        '--includeKey',
        nargs='+',
        default=None,
        help='When specified key(s) to include from the metadata. Matches any key starting with the value. May not be used with excludeKey.',
    )
    metadata.add_argument(
        '--excludeKey',
        nargs='+',
        default=None,
        help='When specified key(s) to exclude from the metadata. Matches any key starting with the value. May not be used with includeKey.',
    )
    metadata.add_argument(
        '--expandSubWorkflows',
        default=False,
        help='When true, metadata for sub workflows will be fetched and inserted automatically in the metadata response.',
    )

    # query arguments
    # TODO: implement CLI entry for query API.

    # group all of the arguments
    args = vars(main_parser.parse_args(arguments))

    # TODO: see if this can be moved or if the commands can be populated from above
    if args['command'] in (
        'submit',
        'wait',
        'status',
        'abort',
        'release_hold',
        'health',
        'metadata',
    ):
        auth_arg_dict = {k: args.get(k) for k in auth_args.keys()}
        auth = CromwellAuth.harmonize_credentials(**auth_arg_dict)
        args['auth'] = auth
        for k in auth_args:
            if k in args:
                del args[k]
    command = getattr(CromwellAPI, args['command'])
    del args['command']
    return command, args


# this should just getattr from CromwellAPI and call the func with args.
# TODO: refactor this module into class-based parsers
def main(arguments=None):
    command, args = parser(arguments)
    API_result = command(**args)
    if isinstance(API_result, requests.Response):
        print(API_result.text)
    else:
        print(API_result)



import os
import json
from google.cloud import storage
from google.oauth2 import service_account


# ask about optional arguments
def get_benchmark(
    output_dir,
    parent_workflow_name,
    cromwell_id,
    google_bucket,
    project_id,
    credentials=None,
):
    """
    Saves a monitoring log for the tasks being called by your pipeline as a combined file (JSON)

    :param output_dir: the desired output directory to store the monitoring log, plots and json file
    :param parent_workflow_name: the name of the workflow being run
    :param cromwell_id: the cromwell id for the workflow being run
    :param google_bucket: the name of google bucket where your workflow was run
    :param project_id: The google project id for the workflow
    :param credentials: your credentials for google as a JSON file (optional
    """

    calls = get_calls(
        google_bucket, project_id, parent_workflow_name, cromwell_id, credentials
    )

    call_logs = get_call_logs(calls)

    file_name = get_path(
        parent_workflow_name + "_" + cromwell_id + "_benchmark_logs.json", output_dir
    )
    store_benchmark(call_logs, file_name)


def get_calls(
    google_bucket, project_id, parent_workflow_name, cromwell_id, credentials=None
):
    """
    :param google_bucket: the name of google bucket where your workflow was run
    :param project_id: The google project id for the workflow
    :param parent_workflow_name: the name of the workflow being run
    :param cromwell_id: the cromwell id for the workflow being run
    :param credentials: your credentials for google as a JSON file
    :return: a list of every call for the monitoring logs
    """

    # if the user does not provide credentials
    if credentials is not None:
        cred = service_account.Credentials.from_service_account_file(credentials)
        storage_client = storage.Client(project_id, credentials=cred)

    else:
        storage_client = storage.Client(project_id)

    # get the appropriate bucket and then task calls of the workflow/cromwell id
    bucket = storage_client.get_bucket(google_bucket)
    blobs = bucket.list_blobs(prefix=parent_workflow_name + "/" + cromwell_id)

    calls = []
    for blob in blobs:
        # if the task call contains a monitoring log
        if "monitoring.log" in blob.name:
            calls.append(blob)

    # return the task calls
    return calls


def get_call_logs(calls):
    """
    :param calls: a list of every call that contians a monitoring log
    :return: returns the information/log of every task call
    """

    call_logs = {}
    for call in calls:
        # get the information of that call
        call_name = os.path.dirname(call.name)
        call_info = call.download_as_string()
        call_logs[call_name] = call_info

    # returns the information/log of every task call
    return call_logs


def get_path(filename, output_dir):
    """
    :param filename: the filename to be saved in the output directory
    :param output_dir: the desired output directory to store the monitoring log, plots and json file
    :return: returns the combined path of file name and  output directory
    """

    # get the combined path of output dir path, filename and sub dir path
    path = os.path.join(output_dir, filename)

    # if the combined path exist
    if os.path.isfile(path):
        # delete that current path
        os.remove(path)

    # return combined path
    return path


def store_benchmark(call_logs, file_name):
    """
    dumps the information of each call task in dictionary format into one JSON file

    :param call_logs: the information of each call task in dictionary format
    :param file_name: name of the file with absolute path to dump the dictionay into
    """

    with open(file_name, "w") as logs:
        json.dump(call_logs, logs, sort_keys=False, indent=4)



from setuptools import setup


CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Natural Language :: English",
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
]

install_requires = [
    'requests>=2.20.0,<3',
    'six>=1.11.0',
    'google-auth>=1.6.1,<2',
    'setuptools_scm>=3.1.0,<4',
]

extras_require = {
    'test': [
        'black==19.3b0',
        'flake8==3.7.7',
        'mock>=2.0.0',
        'requests_mock>=1.4.0',
        'pre-commit==1.14.4',
        'pytest-cov>=2.5.1',
        'pytest>=3.6.3',
        'pytest-timeout>=1.3.1',
    ]
}

setup(
    name='cromwell-tools',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description='Utilities for interacting with the Cromwell workflow engine',
    classifiers=CLASSIFIERS,
    url='http://github.com/broadinstitute/cromwell-tools',
    author='Mint Team',
    author_email='mintteam@broadinstitute.org',
    license='BSD 3-clause "New" or "Revised" License',
    packages=['cromwell_tools'],
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={'console_scripts': ['cromwell-tools = cromwell_tools.cli:main']},
    include_package_data=True,
)



"""
The pkg_resources module will try to get the right version through the setuptools_scm, which will try to
detect the version of cromwell_tools package from any git tags, commit hash codes. This works if this Python package
is either installed from PyPI or installed via git directly.
"""

from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

# By using the below import statement when you call import cromwell_tools you get:
# cromwell_tools.api.status
# cromwell_tools.api.metadata
# cromwell_tools.api.run
# ...
from cromwell_tools.cromwell_api import CromwellAPI as api  # noqa
